"""
:mod:`edx_ace.message` contains the core :class:`Message` and :class:`MessageType`
classes, which allow specification of the content to be delivered by ACE.
"""
import logging
from abc import ABCMeta
from uuid import UUID, uuid4

import attr

from django.apps import apps

import edx_ace.utils.date as date
from edx_ace.monitoring import report as monitoring_report
from edx_ace.recipient import Recipient
from edx_ace.serialization import MessageAttributeSerializationMixin


@attr.s
class Message(MessageAttributeSerializationMixin, metaclass=ABCMeta):
    """
    A ``Message`` is the core piece of data that is passed into ACE.
    It captures the message, recipient, and all context needed to render
    the message for delivery.

    Arguments:
        app_label (str): The name of the Django app that is sending
            this message. Used to look up the appropriate template
            during rendering. Required.
        name (str): The name of this type of message. Used to look up
            the appropriate template during rendering. Required.

        recipient (:class:`.Recipient`): The intended recipient of the
            message. Optional.
        expiration_time(:class:`~datetime.datetime`): The date and time
            at which this message expires. After this time, the message
            should not be delivered. Optional.
        context (dict): A dictionary to be supplied to the template at
            render time as the context.
        send_uuid (:class:`uuid.UUID`): The :class:`uuid.UUID` assigned
            to this bulk-send of many messages.
        language (str): The language the message should be rendered in.
            Optional.
    """

    # mandatory attributes
    # Name is the unique identifier for the message type.
    # Used for:
    #   tracking
    #   discovering message presentation templates
    app_label = attr.ib()
    name = attr.ib()

    # optional attributes
    recipient = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(Recipient)),
    )

    expiration_time = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(date.datetime)),
    )

    context = attr.ib()

    # TODO(later): better naming to distinguish between these 2 UUIDs
    uuid = attr.ib(
        init=False,
        validator=attr.validators.instance_of(UUID),
    )
    send_uuid = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(UUID)),
        default=None
    )
    options = attr.ib()
    language = attr.ib(default=None)
    log_level = attr.ib(default=None)

    @context.default
    def default_context_value(self):
        return {}

    @options.default
    def default_options_value(self):
        return {}

    @uuid.default
    def generate_uuid(self):
        return uuid4()

    @property
    def unique_name(self):
        """
        A unique name for this message, used for logging and reporting.

        Returns: str
        """
        return '.'.join([self.app_label, self.name])

    @property
    def log_id(self):
        """
        The identity of this message for logging.
        """
        return '.'.join([
            self.unique_name,
            str(self.send_uuid) if self.send_uuid else 'no_send_uuid',
            str(self.uuid)
        ])

    def get_message_specific_logger(self, logger):
        """
        Arguments:
            logger (:class:`logging.Logger`): The logger to be adapted.

        Returns:  :class:`.MessageLoggingAdapter` that is specific to this message.
        """
        return MessageLoggingAdapter(logger, {'message': self})

    def report_basics(self):
        monitoring_report('message_name', self.unique_name)
        monitoring_report('language', self.language)

    def report(self, key, value):
        monitoring_report(key, value)


class MessageLoggingAdapter(logging.LoggerAdapter):
    """
    A :class:`logging.LoggingAdapter` that prefixes log items with
    a message :attr:`.log_id`.ABCMeta

    Expects a ``message`` key in its ``extra`` argument which should
    contain the :class:`.Message` being logged for.
    """
    def process(self, msg, kwargs):
        return '[%s] %s' % (self.extra['message'].log_id, msg), kwargs

    def debug(self, msg, *args, **kwargs):
        log_level = self.extra['message'].log_level
        if log_level and log_level <= logging.DEBUG:
            self.info(msg, *args, **kwargs)
        else:
            super().debug(msg, *args, **kwargs)


@attr.s(eq=False, order=False)
class MessageType(MessageAttributeSerializationMixin):
    """
    A class representing a type of :class:`Message`. An instance of
    a ``MessageType`` is used for each batch send of messages.

    Arguments:
        context (dict): Context to be supplied to all messages sent in
            this batch of messages.
        expiration_time (:class:`datetime.datetime`): The time at which
            these messages expire.
        app_label (str): Override the Django app that is used to resolve
            the template for rendering. Defaults to :attr:`.APP_LABEL` or
            to the app that the message type was defined in.
        name (str): Override the message name that is used to resolve
            the template for rendering. Defaults to :attr:`.NAME` or
            to the name of the class.
    """
    NAME = None
    APP_LABEL = None

    context = attr.ib()
    uuid = attr.ib(
        init=False,
        validator=attr.validators.instance_of(UUID),
    )
    expiration_time = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(date.datetime)),
    )
    app_label = attr.ib()
    name = attr.ib()
    options = attr.ib()
    log_level = attr.ib(default=None)

    @context.default
    def default_context_value(self):
        return {}

    @options.default
    def default_options_value(self):
        return {}  # pragma: no cover

    @uuid.default
    def generate_uuid(self):
        return uuid4()

    @name.default
    def default_name(self):
        """ Return default class name. """
        if self.NAME is None:
            return self.__class__.__name__.lower()
        else:
            return self.NAME

    @app_label.default
    def default_app_label(self):
        """ Get default app Label. """
        if self.APP_LABEL is None:
            return apps.get_containing_app_config(self.__class__.__module__).label
        else:
            return self.APP_LABEL

    def personalize(self, recipient, language, user_context):
        """
        Personalize this `MessageType` to a specific recipient, in order to
        send a specific message.

        Arguments:
            recipient (:class:`.Recipient`): The intended recipient of the
                message. Optional.
            language (str): The language the message should be rendered in.
                Optional.
            user_context (dict): A dictionary containing recipient-specific
                context to be supplied to the template at render time.

        Returns: A new :class:`.Message` that has been personalized to a
            specific recipient.
        """
        context = dict(self.context)
        context.update(user_context)
        return Message(
            app_label=self.app_label,
            name=self.name,
            expiration_time=self.expiration_time,
            context=context,
            send_uuid=self.uuid,
            recipient=recipient,
            language=language,
            log_level=self.log_level,
            options=self.options,
        )

    # We override these so that a subtype of MessageType can compare equal
    # to a deserialized non-subtyped MessageType
    def __eq__(self, other):
        if isinstance(other, MessageType):
            return attr.astuple(self) == attr.astuple(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(attr.astuple(self))

    def __ne__(self, other):
        result = self == other
        if result == NotImplemented:
            return NotImplemented
        else:
            return not result



"""
ACE (Automated Communications Engine) is a framework for automatically
sending messages to users.

:mod:`edx_ace` exports the typical set of functions and classes needed to use
ACE.
"""
# -*- coding: utf-8 -*-
from .ace import send
from .channel import Channel, ChannelType
from .message import Message, MessageType
from .policy import Policy, PolicyResult
from .recipient import Recipient
from .recipient_resolver import RecipientResolver

__version__ = '1.0.1'

default_app_config = 'edx_ace.apps.EdxAceConfig'

__all__ = [
    'send',
    '__version__',
    'Message',
    'MessageType',
    'Recipient',
    'RecipientResolver',
    'ChannelType',
    'Channel',
    'Policy',
    'PolicyResult',
]



# pylint: disable=missing-docstring
from unittest.mock import patch

import ddt
from edx_toggles.toggles.testutils import override_waffle_flag

from django.test import TestCase, override_settings

from edx_ace.channel.sailthru import BRAZE_ROLLOUT_FLAG, SailthruEmailChannel
from edx_ace.delivery import deliver
from edx_ace.message import Message
from edx_ace.presentation import render
from edx_ace.recipient import Recipient


@ddt.ddt
class TestSailthruChannel(TestCase):

    def setUp(self):
        self.channel = SailthruEmailChannel()

    def test_render_email_with_sailthru(self):
        message = Message(
            app_label='testapp',
            name='testmessage',
            options={},
            recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
        )

        rendered_email = render(self.channel, message)

        assert '{beacon_src}' in rendered_email.body_html
        assert '{view_url}' in rendered_email.body_html
        assert '{optout_confirm_url}' in rendered_email.body_html

    @override_settings(
        ACE_CHANNEL_SAILTHRU_DEBUG=False,
    )
    @ddt.data(
        ({'from_address': 'custom@example.com'}, {'behalf_email': 'custom@example.com'}),
        ({'reply_to': 'student@example.com'}, {'behalf_email': 'student@example.com'}),
        ({'irrelevant': 'test'}, {}),
        ({}, {}),
    )
    @ddt.unpack
    def test_on_behalf_option_with_sailthru(self, message_options, expected_options):
        """
        Tests sailthru send API is called with on_behalf option
        """
        message = Message(
            app_label='testapp',
            name='testmessage',
            options=message_options,
            recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
        )
        rendered_email = render(self.channel, message)

        with patch('edx_ace.channel.sailthru.SailthruClient.send') as mock_send:
            deliver(self.channel, rendered_email, message)
            self.assertEqual(mock_send.call_args_list[0][1]['options'], expected_options)

    @override_settings(
        ACE_CHANNEL_BRAZE_API_KEY='test-api-key',
        ACE_CHANNEL_BRAZE_APP_ID='test-app-id',
        ACE_CHANNEL_BRAZE_REST_ENDPOINT='rest.braze.com',
        ACE_CHANNEL_SAILTHRU_DEBUG=False,
    )
    @ddt.data(True, False)
    def test_braze_rollout_flag(self, enabled):
        message = Message(
            app_label='testapp',
            name='testmessage',
            recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
        )
        rendered_email = render(self.channel, message)

        with patch('edx_ace.channel.sailthru.SailthruClient.send') as mock_send:
            with patch('edx_ace.channel.sailthru.BrazeEmailChannel') as mock_braze:
                with override_waffle_flag(BRAZE_ROLLOUT_FLAG, active=enabled):
                    deliver(self.channel, rendered_email, message)
                    assert mock_braze.call_count == (1 if enabled else 0)
                    assert mock_send.call_count == (0 if enabled else 1)

    @override_settings(
        ACE_CHANNEL_BRAZE_API_KEY='test-api-key',
        ACE_CHANNEL_BRAZE_APP_ID='test-app-id',
        ACE_CHANNEL_BRAZE_REST_ENDPOINT='rest.braze.com',
        ACE_CHANNEL_SAILTHRU_DEBUG=False,
    )
    @override_waffle_flag(BRAZE_ROLLOUT_FLAG, active=True)
    def test_braze_rollout_properties(self):
        assert self.channel.action_links == [('{{${set_user_to_unsubscribed_url}}}', 'Unsubscribe from this list')]
        assert self.channel.tracker_image_sources == []



# pylint: disable=missing-docstring
from smtplib import SMTPException
from unittest.mock import Mock, patch

from django.core import mail
from django.core.files.temp import NamedTemporaryFile
from django.test import TestCase

from edx_ace.channel.file import PATH_OVERRIDE_KEY, FileEmailChannel
from edx_ace.errors import FatalChannelDeliveryError
from edx_ace.message import Message
from edx_ace.recipient import Recipient
from edx_ace.renderers import RenderedEmail


class TestFilesEmailChannel(TestCase):
    def test_enabled_method(self):
        channel = FileEmailChannel()
        assert channel.enabled()

    def test_happy_path(self):
        channel = FileEmailChannel()
        rendered_message = RenderedEmail(
            from_name='Robot',
            head_html='<img src="https://tracker/.img" />',
            subject='\n Hello from  \r\nRobot ! \n',
            body='Just trying to see what is like to talk to a human!',
            body_html="""
                <p>Just trying to see what is like to talk to a human!</p>

                <hr />
            """,
        )

        with NamedTemporaryFile('r+') as f:
            message = Message(
                app_label='testapp',
                name='testmessage',
                options={
                    'from_address': 'bulk@example.com',
                    PATH_OVERRIDE_KEY: f.name,
                },
                recipient=Recipient(lms_user_id=123, email_address='mr@robot.io'),
            )

            channel.deliver(message, rendered_message)

            contents = f.read()
            assert 'subject: Hello from' in contents
            assert 'to: mr@robot.io' in contents
            assert 'body: Just trying to see what' in contents
            assert 'talk to a human!</p>' in contents



"""
Tests of :mod:`edx_ace.policy`.
"""
from collections import namedtuple
from unittest import TestCase, mock

import ddt

from edx_ace import policy
from edx_ace.channel import ChannelType
from edx_ace.test_utils import StubPolicy


@ddt.ddt
class TestPolicy(TestCase):
    """ Test Policies. """
    PolicyCase = namedtuple('PolicyCase', 'deny_values, expected_channels')

    @ddt.data(
        # allow all
        PolicyCase(deny_values=[set()], expected_channels=set(ChannelType)),

        # deny all
        PolicyCase(deny_values=[set(ChannelType)], expected_channels=set()),

        # deny only email
        PolicyCase(deny_values=[{ChannelType.EMAIL}], expected_channels={ChannelType.PUSH}),  # single policy
        PolicyCase(deny_values=[{ChannelType.EMAIL}, set()], expected_channels={ChannelType.PUSH}),  # multiple policies

        # deny both email and push
        PolicyCase(deny_values=[{ChannelType.EMAIL, ChannelType.PUSH}], expected_channels=set()),  # single policy
        PolicyCase(deny_values=[{ChannelType.EMAIL}, {ChannelType.PUSH}], expected_channels=set()),  # multiple policies

        # deny all and email
        PolicyCase(deny_values=[{ChannelType.EMAIL}, set(ChannelType)], expected_channels=set()),
    )
    @ddt.unpack
    def test_policies(self, deny_values, expected_channels):
        policies = [StubPolicy(deny_value) for deny_value in deny_values]
        with mock.patch.object(policy, 'policies', return_value=policies):
            channels = policy.channels_for(message=mock.Mock())
            self.assertEqual(channels, expected_channels)



"""
:mod:`edx_ace.renderers` contains the classes used by ACE to
render messages for particular types of delivery channels. Each
:class:`ChannelType` has a distinct subclass of :class:`AbstractRenderer`
associated with it, which is used to render messages for all
:class:`Channel` subclasses of that type.
"""
import attr

from django.template import loader


class AbstractRenderer:
    """
    Base class for message renderers.

    A message renderer is responsible for taking one, or more, templates,
    and context, and outputting a rendered message for a specific message
    channel (e.g. email, SMS, push notification).
    """
    rendered_message_cls = None

    def render(self, channel, message):
        """
        Renders the given message.

        Args:
             channel (:class:`Channel`): The channel to render the message for.
             message (:class:`Message`): The message being rendered.

         Returns:
             dict: Mapping of template names/types to rendered text.
        """
        rendered = {}
        for attribute in attr.fields(self.rendered_message_cls):
            # TODO(later): Add comments to explain this difference in
            # behavior between html and txt files, or make it consistent.
            field = attribute.name
            if field.endswith('_html'):
                filename = field.replace('_html', '.html')
            else:
                filename = field + '.txt'
            template = self.get_template_for_message(channel, message, filename)
            render_context = {
                'message': message,
                'channel': channel,
            }
            render_context.update(message.context)
            rendered[field] = template.render(render_context)

        return self.rendered_message_cls(**rendered)  # pylint: disable=not-callable

    def get_template_for_message(self, channel, message, filename):
        """
        Arguments:
            message (:class:`Message`): The message being rendered.
            filename (str): The basename of the template file to look up.

        Returns:
            The full template path to the template to render.
        """
        template_path = '{app_label}/edx_ace/{name}/{channel_type}/{filename}'.format(
            app_label=message.app_label,
            name=message.name,
            channel_type=channel.channel_type.value,
            filename=filename,
        )
        return loader.get_template(template_path)


@attr.s
class RenderedEmail:
    """
    Encapsulates all values needed to send a :class:`.Message`
    over an :attr:`.ChannelType.EMAIL`.
    """
    from_name = attr.ib()
    subject = attr.ib()
    body_html = attr.ib()
    head_html = attr.ib()
    body = attr.ib()


class EmailRenderer(AbstractRenderer):
    """
    A renderer for :attr:`.ChannelType.EMAIL` channels.
    """
    rendered_message_cls = RenderedEmail
